/* Copyright (C) 2009-2010 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * If the integration of miniscope code requires the use of a license
 * other than GPL, an alternative license may be arranged.
 */

#include "Scope.h"
#include "..\miniscope\device\MiniscopeDevice.h"
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <algorithm>

#include "Mutex.h"
#include "ScopedLock.h"
#include "StaticCheck.h"


Mutex mutex;

DWORD WINAPI ThreadProc(LPVOID threaddata)
{
	Scope *scope = reinterpret_cast<Scope*>(threaddata);
	scope->Handle();
    return 0;
}

float Scope::fSampl[] = {
	1553e-6 / SAMPLES_COUNT,
	2397e-6 / SAMPLES_COUNT,
	4148e-6 / SAMPLES_COUNT,
	8267e-6 / SAMPLES_COUNT,
	16470e-6 / SAMPLES_COUNT,
	40944e-6 / SAMPLES_COUNT,
	81835e-6 / SAMPLES_COUNT,
	163965e-6 / SAMPLES_COUNT,
	409845e-6 / SAMPLES_COUNT,
	818966e-6 / SAMPLES_COUNT,
	1639258e-6 / SAMPLES_COUNT,
	4098052e-6 / SAMPLES_COUNT,
};

enum { MAX_VAL = 2048 };

float Scope::fSens[] = {
	0.05 / MAX_VAL,
	0.1 / MAX_VAL,
	0.25 / MAX_VAL,

	0.5	/ MAX_VAL,
	1.0 / MAX_VAL,
	2.5 / MAX_VAL,

	5.0 / MAX_VAL,
	10.0 / MAX_VAL,
	25.0 / MAX_VAL
};

Scope::Scope():
	bRunning(false),
	comPort("COM1"),
    mode(TRIGGER_MODE_AUTO),
    bManualTriggerRequest(false),
    frame_position(0),
    Thread(NULL),
	hCommDev(INVALID_HANDLE_VALUE),
	linePos(0),
	exited(true),
	frameSamplesStarted(false),
	frameSamplesCnt(0)
{
    callbackData = NULL;
    callbackLog = NULL;
	callbackConnect = NULL;
	callbackTrigger = NULL;
	callbackUpdateTimebase = NULL;
	callbackUpdateSensitivity = NULL;
	callbackUpdateCoupling = NULL;

    OutputDebugStr("DSO138 scope object created");
}

Scope::~Scope()
{
    Disconnect();
}

void Scope::Handle(void)
{
	char rxBuf[1024];
    log("RX thread created");
    while (connected)
    {
        // get samples from device
        DWORD readed;
		bool status = ReadFile(hCommDev, rxBuf, sizeof(rxBuf), &readed, NULL);
		if (status == false)
		{
            connected = false;
			if (connected && callbackConnect)
                callbackConnect(DEVICE_DISCONNECTED, (char*)"ReadFile failed, disconnected");
			log("ReadFile error, RX thread exit");
            return;
		}
		OnDataReceive(rxBuf, readed);
	}
    if (hCommDev != INVALID_HANDLE_VALUE)
    {
        CloseHandle(hCommDev);
    }
	hCommDev = INVALID_HANDLE_VALUE;
	log("RX thread exit");
	exited = true;
}

void Scope::log(const char* const lpData, ...)
{
    static char buf[1024]; //determines max message length
    va_list ap;
    va_start(ap, lpData);
    vsnprintf(buf, sizeof(buf), lpData, ap);
    va_end(ap);
    buf[sizeof(buf)-1] = 0;

    if (callbackLog)
        callbackLog(buf);
}

int Scope::Connect(void)
{
    if (hCommDev != INVALID_HANDLE_VALUE)
        CloseHandle(hCommDev);
    hCommDev = INVALID_HANDLE_VALUE;

    COMMTIMEOUTS ctmoNew, ctmoOld;
    char szComPort[32];
    sprintf(szComPort, "\\\\.\\%s", comPort.c_str());
    hCommDev = CreateFile(szComPort, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, 0);
    if (hCommDev != INVALID_HANDLE_VALUE)
    {
        //ustawienia portu
        if (SetupComm(hCommDev, 8192, 4096) == false)
        {
            log("SetupComm call failed!");
        }
        //ustawienia timeoutow
		GetCommTimeouts(hCommDev,&ctmoOld);
		//maximum time, in milliseconds, allowed to elapse between the arrival of two characters
		ctmoNew.ReadIntervalTimeout = 100;
		ctmoNew.ReadTotalTimeoutConstant = 40; //czas dodawany dla calej operacji
		ctmoNew.ReadTotalTimeoutMultiplier = 1;   //to jest mnozone przez liczbe bajtow dla operacji IO
        ctmoNew.WriteTotalTimeoutMultiplier = 0;
        ctmoNew.WriteTotalTimeoutConstant = 600;
        SetCommTimeouts(hCommDev, &ctmoNew);

        // -- ustawienia flag sterujacych DCB --
        dcb.DCBlength = sizeof(dcb);
        GetCommState(hCommDev, &dcb);
        dcb.BaudRate = CBR_115200;
        //jaka i czy jest parzystosc
        dcb.Parity = NOPARITY;
        //liczba bitow stopu
        dcb.StopBits = ONESTOPBIT;
        //wybor liczby bitow danych
        dcb.ByteSize = 8;
        SetCommState(hCommDev, &dcb);
		log( "Opened port %s.", szComPort );
		FlushFileBuffers(hCommDev);

		ConnectCb(DEVICE_CONNECTED, (char*)"DSO138 (DLO138)");
		connected = true;
		exited = false;
		DWORD dwtid;
		if (!Thread)
			Thread = CreateThread(NULL, 0, ThreadProc, this, 0, &dwtid);
		SetThreadPriority(Thread, THREAD_PRIORITY_TIME_CRITICAL);

		frameSamplesStarted = false;
		frameSamplesCnt = 0;
		timebaseIndex = -1;
		sensitivityIndex = -1;
		voltageMultiplier = 1.0f;
		return 0;
	}
    else
    {
        ConnectCb(DEVICE_DISCONNECTED, (char*)"Could not open port. Check Settings/Device Settings.");
        log("Failed opening %s. Change Settings/Device Settings and reconnect.", szComPort);
        return 1;
	}
}

int Scope::Disconnect(void)
{
    // wait for read thread to terminate
	connected = false;
	while (!exited)
	{
		Sleep(100);
	}
	Thread = NULL;
	ConnectCb(DEVICE_DISCONNECTED, (char*)"DSO138 (DLO138) disconnected");
    return 0;
}

int Scope::ManualTrigger(void)
{
    bManualTriggerRequest = true;
    return 0;
}

int Scope::SetTriggerMode(enum E_TRIGGER_MODE m)
{
    mode = m;
    return 0;
}

int Scope::SetTriggerType(enum E_TRIGGER_TYPE type)
{
    log("Trigger type set to %d", type);
    return 0;
}

int Scope::SetTriggerLevel(int level)
{
    //if (level < 0 || level > 1024)
    //    return -1;
    log("Trigger level set to %d", level);
    return 0;
}

int Scope::EnableChannel(int channel, bool state)
{
	ScopedLock<Mutex> lock(mutex);
	pChannelEnable[channel] = state;
	return 0;
}

int Scope::Run(bool state)
{
    bRunning = state;
    if (!bRunning)
    {
        bManualTriggerRequest = false;
        frame_position = 0;
    }
    return 0;
}

void Scope::OnDataReceive(const char* buf, unsigned int count)
{
	for (unsigned int i=0; i<count; i++)
	{
		const char c = buf[i];
		if (c == '\r' || c == '\n')
		{
			if (linePos == 0)
			{
            	// skipping empty line or \n in \r\n sequence
			}
			else
			{
				if (linePos < sizeof(line)-1)
				{
					line[linePos++] = '\0';
					OnLineReceive(line);
					linePos = 0;
				}
			}
		}
		else
		{
			if (linePos < sizeof(line)-1)
			{
				line[linePos++] = c;
			}
		}
	}
}

/*
Net sampling time (us): 40944
Per Sample (us): 19.99
Timebase: 0.5 mS/div
Actual Timebase (us): 499.80
CH1 Coupling: DC, Range: 2V/div
CH2 Coupling: --, Range: +-2048
Triggered: YES
CH1 Stats (V):
	Vmax: 4.98, Vmin: 1.42, Vavr: 3.21, Vpp: 3.56, Vrms: 3.64
	Freq: 1000.39, Cycle: 1.00 ms, PW: 0.50 ms, Duty: 50.00 %
*/
void Scope::OnLineReceive(const char* text)
{
	//log("line = %s", text);
	const char* sampleTimeNeedle = "Net sampling time (us): ";
	if (strncmp(text, sampleTimeNeedle, strlen(sampleTimeNeedle)) == 0)
	{
		float tmp = 1e-6 * SAMPLES_COUNT;
		if (sscanf(text + strlen(sampleTimeNeedle), "%f", &tmp) == 1)
		{
			// find closest sampling time index
			tmp /= 1000000.0f; 	// unit = us
			int minDeltaIndex = -1;
			float minDelta = 1000;
			for (unsigned int i=0; i<sizeof(fSampl)/sizeof(fSampl[0]); i++)
			{
				float delta = fabs(tmp/SAMPLES_COUNT - fSampl[i]) / fSampl[i];
				if (delta < minDelta)
				{
					minDelta = delta;
					minDeltaIndex = i;
				}
			}
			if (minDeltaIndex != timebaseIndex)
			{
				log("samplingTime index = %d", minDeltaIndex);
				timebaseIndex = minDeltaIndex;
				callbackUpdateTimebase(timebaseIndex);
			}
		}
		else
		{
			log("Failed to interpret sample time line");
		}
		frameSamplesStarted = false;		
		return;
	}

	if (!frameSamplesStarted)
	{
		log("line = %s", text);
		const char* dataStartNeedle = "Time\t";
		const char* sensitivityNeedle = "CH1 Coupling: ";
		const char* ch1StatsNeedle = "CH1 Stats";	// needed to get unit for channel: mV or V
		const char* rangeNeedle = "Range: ";
		if (strncmp(text, dataStartNeedle, strlen(dataStartNeedle)) == 0)
		{
			log("start line found");
			callbackTrigger();
			frameSamplesCnt = 0;
			frameSamplesStarted = true;
		}
		else if (strncmp(text, sensitivityNeedle, strlen(sensitivityNeedle)) == 0)
		{
			const char* ptr = text + strlen(sensitivityNeedle);
			if (strncmp(ptr, "DC", 2) == 0)
			{
				callbackUpdateCoupling(0, COUPLING_DC);
			}
			else if (strncmp(ptr, "AC", 2) == 0)
			{
				callbackUpdateCoupling(0, COUPLING_AC);
			}
			else if (strncmp(ptr, "GND", 3) == 0)
			{
				callbackUpdateCoupling(0, COUPLING_GND);
			}
			const char* range = strstr(ptr, rangeNeedle);
			if (range)
			{
				range += strlen(rangeNeedle);
				float tmp;
				if (sscanf(range, "%f", &tmp) == 1)
				{
					if (strstr(range, "mV"))
					{
                    	tmp /= 1000;
					}
					log("Range = %f", tmp);

					int minDeltaIndex = -1;
					float minDelta = 1000;
					for (unsigned int i=0; i<sizeof(fSens)/sizeof(fSens[0]); i++)
					{
						float delta = fabs((tmp * 5 / MAX_VAL) - fSens[i]) / fSens[i];
						if (delta < minDelta)
						{
							minDelta = delta;
							minDeltaIndex = i;
						}
					}
					if (minDeltaIndex != sensitivityIndex)
					{
						log("sensitivityTime index = %d", minDeltaIndex);
						sensitivityIndex = minDeltaIndex;
						callbackUpdateSensitivity(0, sensitivityIndex);
					}
				}
			}
		}
		else if (strncmp(text, ch1StatsNeedle, strlen(ch1StatsNeedle)) == 0)
		{
			if (strstr(text, "(mV)"))
			{
				voltageMultiplier = 0.001f;
			}
			else
			{
				voltageMultiplier = 1.0f;
			}
		}
	}
	else
	{
		/*
		Data format of DLO-138:
		
		Time	CH1	CH2	D_CH1	D_CH2
		0.00	4.90	146	1	0
		19.99	4.88	151	1	0
		39.98	4.90	180	1	0

		CH1 voltage can be either using V or mV
		*/
		float time;
		float analog1;
		int analog2;
		int digital1;
		int digital2;
		if (sscanf(line, "%f %f %d %d %d", &time, &analog1, &analog2, &digital1, &digital2) == 5)
		{
			if (frameSamplesCnt < SAMPLES_COUNT)
			{
				STATIC_CHECK(CHANNELS == 4, ChannelCountMismatch);
				STATIC_CHECK(SAMPLES_COUNT % SAMPLES_PER_CALLBACK_COUNT == 0, SamplesCountMismatch);
				int pos = frameSamplesCnt % SAMPLES_PER_CALLBACK_COUNT;
				int val1;
				if (sensitivityIndex > 0)
				{
					val1 = (analog1 * voltageMultiplier / fSens[sensitivityIndex]) + 0.5f;
					//log("%s\n%f -> %d", line, analog1, val1);
					if (val1 > 32767)
						val1 = 32767;
					else if (val1 < -32768)
						val1 = -32768;
				}
				else
				{
					val1 = 0;
				}
				channelData[0].samples[pos] = val1;
				channelData[1].samples[pos] = analog2;
				channelData[2].samples[pos] = digital1 * 1000;
				channelData[3].samples[pos] = digital2 * 1000;
				frameSamplesCnt++;
				if (pos + 1 == SAMPLES_PER_CALLBACK_COUNT)
				{
					short buf[SAMPLES_PER_CALLBACK_COUNT*CHANNELS];
					int enabledChannelsCnt = 0;
					for (int i=0; i<CHANNELS; i++)
					{
						if (pChannelEnable[i])
						{
							memcpy(buf + (enabledChannelsCnt*SAMPLES_PER_CALLBACK_COUNT), channelData[i].samples, SAMPLES_PER_CALLBACK_COUNT * sizeof(buf[0]));
							enabledChannelsCnt++;
						}
					}
					log("callbackData");
					callbackData(buf, enabledChannelsCnt * SAMPLES_PER_CALLBACK_COUNT * sizeof(buf[0]));
				}
				if (frameSamplesCnt == SAMPLES_COUNT)
				{
					frameSamplesCnt = 0;
					frameSamplesStarted = false;
				}
			}
		}
		else
		{
			log("Failed to interpret data line: %s", line);
		}
	}
}

